
/****************************************************************
*                                                               *
*  IDEINFO: Utility to display IDE drive information            *
*                                                               *
*  Written by Julian Moss.                                      *
*                                                               *
****************************************************************/

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>


#define Identify_Drv		0xEC;
#define Pkt_Identify		0xA1;

#define PrimaryIDEBase		0x1F0;
#define SecondaryIDEBase	0x170;
#define PrimaryHDIRQ		0x76;
#define SecondaryHDIRQ		0x77;

unsigned Int_CtlPort		= 0xA0;
unsigned Int1_CtlPort		= 0x20;
unsigned EOI 				= 0x20;
unsigned HD_Data_Reg;
unsigned HD_DrHd_Reg;
unsigned HD_Cmd_Reg;
unsigned HD_St_Reg;

unsigned char HD_Command;
unsigned char HD_IRQ;
unsigned char Diag_Command = 0x90;

char ltext[4] = "";
unsigned primary = 1;
unsigned cdrom = 0;
unsigned slave = 0;
unsigned dump = 0;
unsigned status;
long timeout = 100000;
long size;

struct IDE_Info						/* DESCRIPTION                         WORD */
{
	unsigned Gen_Config;			/* general configuration                  0 */
	unsigned No_Log_Cyl;			/* no. of logical cylinders               1 */
	unsigned reserved1;
	unsigned No_Log_Hd;				/* no. of logical heads                   3 */
	unsigned No_UnfBPT;				/* no. of unformatted bytes per track     4 */
	unsigned No_UnfBPS;				/* no. of unformatted bytes per sector    5 */
	unsigned No_Log_SPT;			/* no. of logical sectors per track       6 */
	unsigned No_Byt_ISG;			/* no. of bytes in the inter-sector gap   7 */
	unsigned No_Byt_Sync;			/* no. of bytes in the sync fields        8 */
	unsigned No_VendInfo;			/* no. of bytes in vendor unique info     9 */
	char Serial_No[20];				/* serial number (in ASCII)              10 - 19 */
	unsigned Buffer_Type;			/* buffer type                           20 */
	unsigned Buffer_Size;			/* buffer size                           21 */
	unsigned ECC_Bytes_Avl;			/* ECC bytes available                   22 */
	char FirmWare_Rev[8];   		/* firmware revision no.                 23 */
	char Model_No[40];				/* model number                          27 */
	unsigned MultSectTfr;           /* number of sectors per multi-block tfr 47 */
	unsigned DWord_IO;              /* double word I/O supported?            48 */
	unsigned Capabilities;			/* capabilities                          49 */
	unsigned reserved2;
	unsigned PIO_Cycle;             /* PIO Cycle Timing                      51 */
	unsigned DMA_Cycle;             /* DMA Cycle Timing                      52 */
	unsigned Field_Valid;           /* Field Validity                        53 */
	unsigned No_Cur_Cyl;            /* current number of cylinders           54 */
	unsigned No_Cur_Hd;             /* current number of heads               55 */
	unsigned No_Cur_SPT;            /* current number of sectors per track   56 */
	long Cur_Capacity;              /* current capacity (sectors)            57 */
	unsigned MultSectValid;			/* no. of sectors per transfer in use    59 */
	long User_Addr_Sect;            /* number of user-addressible sectors    60 */
	unsigned Singleword_DMA;        /* single word DMA mode                  62 */
	unsigned Multiword_DMA;         /* multi word DMA mode                   63 */
	unsigned Enhanced_PIO;          /* enhanced PIO mode                     64 */
    char spare[448];
};

union IDEDATA
{
	struct	IDE_Info	Data;
	unsigned			Buffer[256];
};

union IDEDATA	IDE_Data;

typedef struct DPT far * DPTPTR;

struct DPT
{
	unsigned I13_Cyls;				/* no. of cylinders presented at int 13h interface	*/
	unsigned char I13_Heads;		/* no. of heads presented at int 13h interface		*/
	unsigned char Signature;		/* A?h indicates translation used					*/
	unsigned char Drv_SPT;			/* no. of SPT at drive after translation			*/
	unsigned Precomp;
	unsigned char reserved3;
	unsigned char Control;			/* drive control byte;								*/
	unsigned Drv_Cyls;				/* no. of cylinders at drive after translation		*/
	unsigned char Drv_Heads;		/* no. of heads at drive after translation			*/
	unsigned Landing;
	unsigned char I13_SPT;			/* no. of SPT presented at int 13h interface		*/
	unsigned char Checksum;
};

DPTPTR dptpointer;

#define I13Vector ((unsigned long far *) _MK_FP(0x0, 0x104))

short HD_Status;
 
void strxcpy(char dest[], char source[], int num)
{
int i,j;
char c;
	j = 0;
    for ( i=0; i<num; i +=2)
    {
    	c = source[i+1];
    	if ((c != ' ') || j)
    	{
    		dest[j] = c;
    		j++;
    	}
    	c = source[i];
    	if ((c != ' ') || j)
    	{
    		dest[j] = c;
    		j++;
    	}
    }
    dest[j] = 0x0;
}

char *logical(unsigned val)
{
    if ( val )
        strcpy(ltext,"Yes");
    else
        strcpy(ltext,"No ");
    return ltext;
}

int readideinfo()
{
int i;
long t;
short HD_Status;
	t = 0;
	do
	{
		HD_Status = _inp(HD_St_Reg);
		t++;
	}
	while ((HD_Status != 0x50) & (t < timeout)); 
	if (HD_Status == 0x50)
	{
		if (slave)
			_outp(HD_DrHd_Reg, 0xB0);
/*		else
			_outp(HD_DrHd_Reg, 0xA0); */
		_outp(HD_Cmd_Reg, HD_Command);
/*		printf("****outp(%4X, %2X)\n",HD_Cmd_Reg, HD_Command); */
		printf("Interrogating Drive ...");
		t = 0;
		do
		{
			HD_Status = _inp(HD_St_Reg);
			t++;
		}
		while ((HD_Status != 0x58) & (t < timeout) );
		printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
		printf("Reading data ...");
		if (t < timeout)
		{
			for (i=0; i<256; i++)
				IDE_Data.Buffer[i] = _inpw(HD_Data_Reg);
		}
		printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
	}
	return HD_Status;	
}

int resetcontroller()
{
long t;
short HD_Status;
	_outp(HD_Cmd_Reg, Diag_Command);
	t = 0;
	do
	{
		HD_Status = _inp(HD_St_Reg);
		t++;
	}
	while ((HD_Status != 0x50) & (t < timeout));
	return HD_Status;
}

void printhelp()
{
	printf("Command line switches:\n\n");
	printf("\t/?\tPrint this help screen.\n");
	printf("\t/1\tDisplay primary IDE channel details (default).\n");
	printf("\t/2\tDisplay secondary IDE channel details.\n");
	printf("\t/c\tDisplay IDE CD-ROM drive details.\n");
	printf("\t/s\tDisplay slave drive details.\n");
	printf("\t/d\tDump information in hex format.\n");
	
	printf("\n");
}

void printidename()
{
char text[48];
	strxcpy(text,IDE_Data.Data.Model_No,40);
	printf("Model : %-40s",text);
	strxcpy(text,IDE_Data.Data.Serial_No,20);
    printf(" Ser. No : %-20s\n",text);
    if (IDE_Data.Data.Gen_Config & 0x80)
    {
    	cdrom = 1;
    	strcpy(text,"(CD-Rom Drive)   ");
    }
    else
    	if (IDE_Data.Data.Gen_Config & 0x80)
    		strcpy(text,"(Removable Media)");
	    else
    		strcpy(text,"(Fixed Media)    ");
    printf("        %17s",text);
    strxcpy(text,IDE_Data.Data.FirmWare_Rev,8);
    printf("             Firmware Revision  : %-10s\n",text);
}


void printdpt()
{
    dptpointer = (DPTPTR) *I13Vector;
    size = (long)dptpointer->I13_Heads*(long)dptpointer->I13_Cyls*(long)dptpointer->I13_SPT / 2048;
   	printf("         BIOS - Heads:%5u  Cylinders:%5u  Sectors/Trk:%5u  (%4LuMB)\n",
	        dptpointer->I13_Heads,dptpointer->I13_Cyls,dptpointer->I13_SPT,size);
	if ((dptpointer->Signature >> 4) == 0x0A)
	{
   	printf("   Translated - Heads:%5u  Cylinders:%5u  Sectors/Trk:%5u\n",
	        dptpointer->Drv_Heads,dptpointer->Drv_Cyls,dptpointer->Drv_SPT);
	}
}

void printidecapabilities()
{
char text[48];
	if (!cdrom)
	{
	    printf("Drive parameters:\n");
	    size = (long)IDE_Data.Data.No_Log_Hd*(long)IDE_Data.Data.No_Log_Cyl*(long)IDE_Data.Data.No_Log_SPT/2048;
    	printf("      Default - Heads:%5u  Cylinders:%5u  Sectors/Trk:%5u  (%4LuMB)\n",
	        IDE_Data.Data.No_Log_Hd,IDE_Data.Data.No_Log_Cyl,IDE_Data.Data.No_Log_SPT,size);
    	if (IDE_Data.Data.No_Cur_Cyl)
	    	printf("      Current - Heads:%5u  Cylinders:%5u  Sectors/Trk:%5u\n",
    	    	IDE_Data.Data.No_Cur_Hd,IDE_Data.Data.No_Cur_Cyl,IDE_Data.Data.No_Cur_SPT);
    	if (primary & (!slave)) printdpt();
    	printf("\n");
	}
	if (IDE_Data.Data.Buffer_Type > 0)
	{
	    printf("Controller Buffer Size             : %4u KB  ( %4u sectors )\n",
    		IDE_Data.Data.Buffer_Size/2,IDE_Data.Data.Buffer_Size);
		if ( IDE_Data.Data.Buffer_Type )
		{
			if ( IDE_Data.Data.Buffer_Type == 1 ) strcpy(text,"Single Ported Single Sector        ");
			if ( IDE_Data.Data.Buffer_Type == 2 ) strcpy(text,"Dual Ported Multi Sector           ");
			if ( IDE_Data.Data.Buffer_Type == 3 ) strcpy(text,"Dual Ported Multi Sector with cache");
			if ( IDE_Data.Data.Buffer_Type > 3 ) strcpy(text,"Unknown                             ");
    		printf("Controller Buffer Type             : %35s\n",text);
    	}
    }
    printf("\n");
    printf("Logical Block Addressing supp. : %3s\n",
    	logical(IDE_Data.Data.Capabilities & 0x0200) );    
    printf("Multi-sector transfers supp.   : %3s",
    	logical(IDE_Data.Data.MultSectTfr & 0x00FF));
    if (IDE_Data.Data.MultSectTfr & 0x00FF)
    {
    	printf("    Maximum : %3u",IDE_Data.Data.MultSectTfr & 0x00FF);
    	if (IDE_Data.Data.MultSectValid & 0x0100)
    		printf("         Current  : %3u",IDE_Data.Data.MultSectValid & 0x00FF);
    }
    printf("\nDouble word I/O capable        : %3s\n",
    	logical(IDE_Data.Data.DWord_IO));
    printf("IORDY operation supported?     : %3s\n",
        logical(IDE_Data.Data.Capabilities&0x0800));
    if ( IDE_Data.Data.Field_Valid & 0x0002 )
    {
        printf("Multiword DMA Mode 1 supported : %3s    Multiword DMA Mode 1 active    : %3s\n",
            logical(IDE_Data.Data.Multiword_DMA & 2),logical(IDE_Data.Data.Multiword_DMA & 512));
    }
        strcpy(text,"0 ");
        if ((IDE_Data.Data.PIO_Cycle >> 8) == 1) strcpy(text,"0 1 ");
        if ((IDE_Data.Data.PIO_Cycle >> 8) == 2) strcpy(text,"0 1 2 ");
        if (IDE_Data.Data.Enhanced_PIO & 0x0001 ) strcat(text,"3 ");
        if (IDE_Data.Data.Enhanced_PIO & 0x0002 ) strcat(text,"4 ");
        printf("PIO Modes supported            : %-8s\n",text);       
}

void dumpideinfo()
{
int i;
	printf("\n");
	for (i=0; i<80; i+=10)
	{
		printf("\t%4X  %4X  %4X  %4X  %4X  %4X  %4X  %4X  %4X  %4X\n",
			IDE_Data.Buffer[i],IDE_Data.Buffer[i+1],IDE_Data.Buffer[i+2],IDE_Data.Buffer[i+3],
			IDE_Data.Buffer[i+4],IDE_Data.Buffer[i+5],IDE_Data.Buffer[i+6],IDE_Data.Buffer[i+7],
			IDE_Data.Buffer[i+8],IDE_Data.Buffer[i+9]);
	}
}

void main(int argc, char *argv[])
{
int c;

	HD_Data_Reg = PrimaryIDEBase;
	HD_DrHd_Reg = HD_Data_Reg + 6;
	HD_Cmd_Reg = HD_Data_Reg + 7;
	HD_St_Reg = HD_Data_Reg + 7;
	HD_Command = Identify_Drv;
	HD_IRQ = PrimaryHDIRQ;
	
    printf("\nIDEINFO ver 1.0 (11 Sep 1994)        Copyright (c) 1994  J M Technical Services\n\n");
    
    /* process command line switches */
    
    if (argc > 1)
    {
    	for (c=1;c<argc;c++)
    	{
    		switch (*(argv[c]+1))
    		{
    			case '?':
    				printhelp();
    				exit(0);
    				break;
    			case '2':
					HD_Data_Reg	= SecondaryIDEBase;
					HD_DrHd_Reg	= HD_Data_Reg + 6;
					HD_Cmd_Reg	= HD_Data_Reg + 0x7;
					HD_St_Reg	= HD_Data_Reg + 0x206;
					HD_IRQ		= SecondaryHDIRQ;
					primary = 0;
    				break;
    			case 'c':
    			case 'C':
    				cdrom = 1;
    				HD_Command = Pkt_Identify;
    				break;
    			case 'd':
    			case 'D':
    				dump = 1;
    				break;
    			case 's':
    			case 'S':
    				slave = 1;
    				break;
    			default:
    				printf("Type IDEINFO /? for command line options.\n\n");
    				exit(0);
    		}
    	}
    }
    
    /* print channel details */
    
    if (primary)
    	printf("Primary IDE Channel (1F0h - 1F7h / 3F6h - 3F7h, IRQ 14)  ");
    else
    	printf("Secondary IDE Channel (170h - 177h / 376h - 377h, IRQ 15)  ");
    	
    if (slave)
    	printf("Slave drive\n\n");
    else
    	printf("Master drive\n\n");    
    
    status = readideinfo();
    
    /* print drive details */
    
    if (!(status & 0x01) | dump)
    {
	    printidename();
    	printf("\n");
    	if (dump)
    		dumpideinfo();
    	else
    		printidecapabilities();
	}
    else
    {
    	if (status == 0xFF)
    		printf("IDE channel disabled or not installed.\n\n");
    	else
    	{
   			printf("Error on IDE channel: status %#4X.\n\n", _inp(HD_St_Reg));
			status = resetcontroller();
		}
    	exit(1);
    }
}
